/* eslint-disable camelcase */
// This file contains the Javascript implementation of a schaakbord.
// It defines the scripts belonging to the UserInterface that is define in schaakbord.html,
// and it used the chessboard as defined in chessboard...js
//
// The main purpose of the functions in this file are to:
// * link actions to buttons
// * allow for addition (and editing) of positions in a list of positions
// * allow for addition (and editing) of moves for a position
// * allow saving the position and moves to a file, and reloading them from
//   file.
//

// ********************************************************************************
// Chessboards
// ********************************************************************************
var config = {
  draggable: true,
  dropOffBoard: 'trash',
  sparePieces: true,
  onChange: onChange
}

// ----------------------------------------------------------------------
// Create chessboard
var board = ChessBoard('board', config)

// ----------------------------------------------------------------------
// Create a dummy chessboard. This board is being used to
// 'play' moves and in that way get the resulting position
// after the moves.
// This is a.o. used for checkAndDoMove()
var dummyboard = ChessBoard('dummy', { })
// The dummy board should NOT be on the screen
document.getElementById('dummy').style.display = 'none'

// -----------------------------------------------------------------------
// Do the onChange event handler?
// This is only for board
var skipOnChange = false

// ----------------------------------------------------------------------
// Wrapper for changing position on the board
// Turns of onChange()
function setPositionOnBoard (position, animate = true) {
  skipOnChange = true
  board.position(position, animate)
  skipOnChange = false
}

// ----------------------------------------------------------------------
// Wrapper for board.clear()
function clearBoard () {
  skipOnChange = true
  board.clear()
  skipOnChange = false
}

// ----------------------------------------------------------------------
// Wrapper for board.start()
function startBoard () {
  skipOnChange = true
  board.start()
  skipOnChange = false
}

// /////////////////////////////////////////////////////////////////////////////////
// Event Handlers
// /////////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------
// Resize the board if the window is resized.
$(window).bind('resize', function (e) {
  // Visibility is not retained after resize (?!)
  var v = board.sparePiecesVisibility()
  board.resize()
  board.sparePiecesVisibility(v)
})

// ----------------------------------------------------------------------
// Go to fullscreen m ode
// https://www.w3schools.com/howto/howto_js_fullscreen.asp
function openFullscreen () {
  var elem = document.documentElement
  if (elem.requestFullscreen) {
    elem.requestFullscreen()
  } else if (elem.mozRequestFullScreen) { /* Firefox */
    elem.mozRequestFullScreen()
  } else if (elem.webkitRequestFullscreen) { /* Chrome, Safari and Opera */
    elem.webkitRequestFullscreen()
  } else if (elem.msRequestFullscreen) { /* IE/Edge */
    elem.msRequestFullscreen()
  }
}

// ----------------------------------------------------------------------
// Leave fullscreen mode
function closeFullscreen () {
  if (document.exitFullscreen) {
    document.exitFullscreen()
  } else if (document.mozCancelFullScreen) { /* Firefox */
    document.mozCancelFullScreen()
  } else if (document.webkitExitFullscreen) { /* Chrome, Safari and Opera */
    document.webkitExitFullscreen()
  } else if (document.msExitFullscreen) { /* IE/Edge */
    document.msExitFullscreen()
  }
}

// ----------------------------------------------------------------------
// Toggles fullscreenmode.
function onToggleFullscreen () {
  if (!window.screenTop && !window.screenY) {
    closeFullscreen()
  } else {
    openFullscreen()
  }
}

// ----------------------------------------------------------------------
// Toggles visibility of the left and right column (but not
// the bottom right buttons).
function onTogglePanels () {
  // Check current state (left column is taken as reference)
  if (document.getElementById('left_col').style.visibility === 'hidden') {
    // adapt visibility...
    document.getElementById('left_col').style.visibility = 'visible'
    document.getElementById('right_col').style.visibility = 'visible'
  } else {
    document.getElementById('left_col').style.visibility = 'hidden'
    document.getElementById('right_col').style.visibility = 'hidden'
  }
}

// ----------------------------------------------------------------------
// General keys for the whole User Interface.
function onKeyPress (event) {
  if (event.key === 'L') { // Hide left pane
    document.getElementById('left_col').style.visibility = 'hidden'
  }
  if (event.key === 'l') { // Show left pane
    document.getElementById('left_col').style.visibility = 'visible'
  }
  if (event.key === 'R') { // Hide right pane
    document.getElementById('right_col').style.visibility = 'hidden'
  }
  if (event.key === 'r') { // Show right pane
    document.getElementById('right_col').style.visibility = 'visible'
  }
  if (event.key === 'B') { // Hide bottom right pane
    document.getElementById('bottom_right').style.visibility = 'hidden'
  }
  if (event.key === 'b') { // Show bottom right pane
    document.getElementById('bottom_right').style.visibility = 'visible'
  }
  if (event.key === 's') { // Start opstelling
    startBoard()
  }
  if (event.key === 'e') { // Leeg veld
    clearBoard()
  }
  if (event.key === 'o') { // Open file
    document.getElementById('loadfile').click()
  }
  if (event.key === 'p') { // Vorige (previous) stelling
    onPreviousPosition()
  }
  if (event.key === 'n') { // Volgende (next) stelling
    onNextPosition()
  }
  if (event.key === 'z') { // Doe een zet
    onNextMove()
  }
  if (event.key === 'Z') { // Zet de laatste zet terug
    onPreviousMove()
  }
  if (event.key === 'A') {
    board.clearAnnotation()
  }
  if (event.key === 'f') {
    board.flip()
  }
  if (event.key === 't') { // Zet de huidige stelling terug, ook de moves
    updatePositionInfoAndBoardAfterPosSelect()
  }
  if (event.key === 'T') { // Zet de huidige stelling terug, met de huidige zet
    updateBoardAfterMoveSelect()
  }
  if (event.key === 'h') {
    window.open('digidemobord-handleiding.html')
  }
}

// ----------------------------------------------------------------------
// If 'Enter' is pressed in the FEN field, the board is updated.
function onKeyPressInputFEN (event) {
  if (event.key === 'Enter') {
    onFENToBoard()
  }
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// When 'Enter' is pressed in the position name, the following is done:
// * Search for the corresponding name.
// * If the name does not exist --> add current position
function onKeyPressPosition (event) {
  if (event.key === 'Enter') {
    var name = document.getElementById('position_name').value
    if (name !== '' && !searchPositionWithName(name)) {
      addPositionToPosList()
      document.getElementById('register_moves').checked = true
      updatePositionInfoAndBoardAfterPosSelect()
      updateUI()
    }
  }
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// When 'Enter' is pressed in the moves text input name, the moves
// of the current position are updated (replaced).
function onKeyPressMoves (event) {
  if (event.key === 'Enter') {
    addMovesToCurrentPosition(true)
    document.getElementById('register_moves').checked = false
    updateBoardAfterMoveSelect()
    updateUI()
  }
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// Adds the pieces (given as text) to the positions on the board.
function doAddPieces (side, pieces) {
  // Replace any punctuation chars with spaces.
  pieces = pieces.replace(/\./g, ' ')
  pieces = pieces.replace(/,/g, ' ')
  pieces = pieces.replace(/;/g, ' ')

  // Get board position
  var position = board.position()
  // Get the pieces...
  pieces = pieces.split(' ')
  // For each piece
  for (var i = 0; i < pieces.length; ++i) {
    // Check the code
    var piece_string = pieces[i].trim()
    if (piece_string !== '') {
      // Get last two characters (i.e. the square)
      var square = piece_string.slice(-2)
      // If the first characters are a valid piece..
      if (validPieceUI(piece_string.slice(0, -2))) {
        // Create the piece code
        var piece = side + pieceUIToBoard[piece_string.slice(0, -2)]
        // If the square and piece are valid...
        if (validSquare(square) && validPieceCode(piece)) {
          // Add to the position...
          position[square] = piece
        }
      }
    }
  }
  // Update the board based on the new position.
  setPositionOnBoard(position)
}

// ----------------------------------------------------------------------
// If enter is pressed on the white pieces text inbox,
// add the pieces to the board.
function onKeyPressWhitePieces (event) {
  if (event.key === 'Enter') {
    // Don't track moves when adding pieces ...
    document.getElementById('register_moves').checked = false
    doAddPieces('w', document.getElementById('white_pieces').value)
    updateUI()
  }
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// If enter is pressed on the black pieces text inbox,
// add the pieces to the board.
function onKeyPressBlackPieces (event) {
  if (event.key === 'Enter') {
    // Don't track moves when adding pieces ...
    document.getElementById('register_moves').checked = false
    doAddPieces('b', document.getElementById('black_pieces').value)
    updateUI()
  }
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// Updates the board based on the current FEN.
function onFENToBoard () {
  var fenstring = document.getElementById('inputfen').value
  document.getElementById('register_moves').checked = false
  setPositionOnBoard(fenstring)
  updateUI()
}

// ----------------------------------------------------------------------
// Update the current FEN based on the current position
// on the board.
function onBoardToFEN () {
  document.getElementById('inputfen').value = board.fen()
}

// ----------------------------------------------------------------------
// Update the white pieces based on the board.
function onBoardToWhitePieces () {
  document.getElementById('white_pieces').value = getPiecesAsString('w')
}

// ----------------------------------------------------------------------
// Update the black pieces based on the board.
function onBoardToBlackPieces () {
  document.getElementById('black_pieces').value = getPiecesAsString('b')
}

// ----------------------------------------------------------------------
// Called when a position is added.
// Only adds a position when the name is not present yet
// in the position list.
function onAddPosition () {
  if (addPositionToPosList()) {
    document.getElementById('register_moves').checked = true
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Called when a position is removed.
// Removes the position with the name given.
function onRemovePosition () {
  if (removePositionFromPosList()) {
    document.getElementById('register_moves').checked = false
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Called when a position is removed.
// Updates the current position (not its moves). If a new name is given,
// is should not match an existing name.
function onUpdatePosition () {
  if (updatePositionInPosList()) {
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Adds the current moves to the current position.
function onAddMoves () {
  if (addMovesToCurrentPosition(true)) {
    document.getElementById('register_moves').checked = false
    updateBoardAfterMoveSelect()
    // updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Removes all moves to the current position.
function onRemoveMoves () {
  if (removeMovesFromCurrentPosition()) {
    updateBoardAfterMoveSelect()
    // updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Replaces the moves of the current position with the new moves.
function onUpdateMoves () {
  if (addMovesToCurrentPosition(false)) {
    document.getElementById('register_moves').checked = false
    updateBoardAfterMoveSelect()
    // updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Update all after a new position has been selected
// (called from the list)
function onNewPositionSelected () {
  document.getElementById('register_moves').checked = false
  updatePositionInfoAndBoardAfterPosSelect()
  updateUI()
}

// ----------------------------------------------------------------------
// Update all after a new move has been selected
// (called from the list)
function onNewMoveSelected () {
  document.getElementById('register_moves').checked = false
  updateBoardAfterMoveSelect()
  updateUI()
}

// ----------------------------------------------------------------------
// Is called when the position on the board changes.
// When the board changes, it will find out the move that has been performed.
// If register_moves is set, a nice text representation is added to the
// 'moves' parameter, and if it is a special move
// (castling, en passent, promotion), the move itself (i.e. the board position
// will be finalized automatically (promotion is always Queen promotion).
function onChange (oldPos, newPos) {
  // skipOnChange should be true if the board is changed from
  // this application. onChange() should only run if there were
  // 'manual' changes on the board.
  if (skipOnChange) return

  // returnpos will contain the new position to be returned.
  // It will be set if the change detected should be 'finalized',
  // such as in the case of castling, a promotion or en passent moves.
  var returnpos = null

  // Find out what has been changed ..
  var move = findMoveBetweenPositions(oldPos, newPos)

  if (move !== '') { // valid move...
    // Use the dummyboard to test the move....
    dummyboard.position(oldPos)

    // Do the move on the dummy board
    var nicemove = checkAndDoMove(dummyboard, move)

    // It could happen the the new position in the dummyboard is
    // different from the newPos given. This is in the following cases:
    // * castling (when king castles, the rook is moved as well)
    // * en passent moves: the fiendly pawn is additionally removed
    // * promotion: a Queen is added to the promotion square. Minor promotion
    //   is not supported yet.

    // Add the move to the text representation
    if (document.getElementById('register_moves').checked) {
      if (document.getElementById('moves').value !== '') { // add space if already filled
        document.getElementById('moves').value += ' '
      }
      document.getElementById('moves').value += nicemove
    }

    // If the position should be updated...
    if (dummyboard.position() !== newPos) {
      // Store the new position
      returnpos = dummyboard.position()
    }
  }

  // Check whether the new position matches the next move ...
  if (returnpos) {
    dummyboard.position(returnpos)
  } else {
    dummyboard.position(newPos)
  }

  // Get fen of the board (after the change), and
  var newfen = dummyboard.fen()
  // get fen of the board as it should be if no changes would be applied.
  var curfen = getFENCurrentPosition(0)

  // If new position does not match the expected one
  if (newfen !== curfen) {
    // Check whether the next move has been manually done ...
    var nextfen = getFENCurrentPosition(1)
    if (newfen === nextfen) {
      // Yes, the change corresponds to the manual move, update move selected
      var moves_list = document.getElementById('moves_list')
      moves_list.selectedIndex = moves_list.selectedIndex + 1
    }
  }

  updateStatusLine(newfen)

  // Notify the board that the new position should be updated
  if (returnpos) return returnpos
}

// ======================================================================
// Status Position Patter Matching

// ----------------------------------------------------------------------
// Prepare the elements (visibility, initialization) that
// allow pattern-matching based search for positions (on names)
function onStatusPositionClicked () {
  // Reset the elements ...
  var pos_name = ''
  var pos_list = document.getElementById('position_list')
  if (pos_list.length > 0 && pos_list.selectedIndex > -1) {
    pos_name = pos_list.options[pos_list.selectedIndex].text
  }

  document.getElementById('search_status_position').value = pos_name
  var match_list = document.getElementById('match_list')
  // Reset list
  match_list.options.length = 0

  // Activate the right elements ...
  document.getElementById('status_position').style.display = 'none'
  document.getElementById('search_status_position').style.display = 'inline-block'
  document.getElementById('match_list').style.display = 'inline-block'
  document.getElementById('search_status_position').focus()

  // Fills the dropdown list with the current matching positions..
  var regexp = RegExp(pos_name)
  for (var i = 0; i < pos_list.length; ++i) {
    if (pos_list.options[i].text.match(regexp)) {
      match_list.add(new Option(pos_list.options[i].text, ''))
    }
  }
}

// ----------------------------------------------------------------------
// Reset the visibility of the elements that allow pattern-matched
// search for positions.
function searchPositionReset () {
  document.getElementById('status_position').style.display = 'inline-block'
  document.getElementById('search_status_position').style.display = 'none'
  document.getElementById('match_list').style.display = 'none'
}

// ----------------------------------------------------------------------
// When an option in the match-list is selected, this option is used.
function onMatchListChanged () {
  var match_list = document.getElementById('match_list')
  if (match_list.length > 0 && match_list.selectedIndex > -1) {
    var name = match_list.options[match_list.selectedIndex].text
    searchPositionWithName(name)
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
  searchPositionReset()
}

// ----------------------------------------------------------------------
// Prevents key propagations for the text input for pattern-matching
// based searching of position names...
function onKeyPressSearchPosition (event) {
  event.stopPropagation()
}

// ----------------------------------------------------------------------
// Keyup event for text input for pattern-matching based
// search of position names. Apparently, key-up is required
// to catch 'Escape' (does not work with keypress ??!!)
function onKeyUpSearchPosition (event) {
  var match_list = document.getElementById('match_list')
  // Depending on the key ...
  switch (event.key) {
    // Enter means: select!
    // The currently selected (or first one) of the dropdown is used
    // If nothing is selected, and nothing in the dropdown, nothing is performed.
    case 'Enter':
      var selected = match_list.selectedIndex
      if (selected < 0 && match_list.length > 0) selected = 0
      if (selected > -1) {
        searchPositionWithName(match_list.options[selected].text)
      }
      searchPositionReset()
      break
    // Escape means: reset and ignore any selectoin
    case 'Escape':
      searchPositionReset()
      break
    // Arrows can be used to move up and down the dropdown selection
    case 'ArrowUp':
      match_list.selectedIndex = match_list.selectedIndex - 1
      if (match_list.selectedIndex < 0) match_list.selectedIndex = 0
      break
    case 'ArrowDown':
      if (match_list.selectedIndex + 1 < match_list.length) {
        match_list.selectedIndex = match_list.selectedIndex + 1
      }
      break
    default:
      // In all other cases,
      // update list with pattern matched results...
      var pos_list = document.getElementById('position_list')
      var pattern = document.getElementById('search_status_position').value
      match_list.options.length = 0
      for (var i = 0; i < pos_list.length; ++i) {
        var regexp = RegExp(pattern)
        if (pos_list.options[i].text.match(regexp)) {
          match_list.add(new Option(pos_list.options[i].text, ''))
        }
      }
  }
  event.stopPropagation()
  updatePositionInfoAndBoardAfterPosSelect()
  updateUI()
}

// ----------------------------------------------------------------------
// Called when the register_moves status is changed.
// Updates the visual representation of the record button.
function onRegisterMovesStatusChanged () {
  var checked = document.getElementById('register_moves').checked
  if (checked) {
    document.getElementById('record').style.color = 'red'
  } else {
    document.getElementById('record').style.color = 'black'
  }
}

// ----------------------------------------------------------------------
// Update (flip) the interaction mode, from pieces to/from annotations.
function onInteractionModeClicked () {
  if (board.interactionMode() === 'annotations') {
    board.interactionMode('pieces')
    document.getElementById('interactionMode').innerHTML = '[&nbsp;]&nbsp;'
  } else {
    board.interactionMode('annotations')
    document.getElementById('interactionMode').innerHTML = '[A]&nbsp;'
  }
}

// ----------------------------------------------------------------------
// Record button is clicked: change the register_moves status
function onRecordClicked () {
  var checked = document.getElementById('register_moves').checked
  checked = !checked
  document.getElementById('register_moves').checked = checked
  updateUI()
}

// ----------------------------------------------------------------------
// Load a file when the status line filename is clicked.
function onStatusFileNameClicked () {
  document.getElementById('loadfile').click()
}

// ----------------------------------------------------------------------
// Move the board (and the position list) to the first position
// If a new position could be selected, the annotations are
// also cleared.
function onFirstPosition () {
  var pos_list = document.getElementById('position_list')
  if (pos_list.length > 0) {
    pos_list.selectedIndex = 0
    board.clearAnnotation()
    document.getElementById('register_moves').checked = false
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board (and the position list) to the previous position, if
// possible.
// If a new position could be selected, the annotations are
// also cleared.
function onPreviousPosition () {
  var pos_list = document.getElementById('position_list')
  if (pos_list.selectedIndex > 0) {
    pos_list.selectedIndex = pos_list.selectedIndex - 1
    board.clearAnnotation()
    document.getElementById('register_moves').checked = false
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board (and the position list) to the next position, if
// possible.
// If a new position could be selected, the annotations are
// also cleared.
function onNextPosition () {
  var pos_list = document.getElementById('position_list')
  if (pos_list.selectedIndex < pos_list.length - 1) {
    pos_list.selectedIndex = pos_list.selectedIndex + 1
    // Will reset all values for the position, and clear
    // annotations.
    board.clearAnnotation()
    document.getElementById('register_moves').checked = false
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board (and the position list) to the last position, if
// possible.
// If a new position could be selected, the annotations are
// also cleared.
function onLastPosition () {
  var pos_list = document.getElementById('position_list')
  if (pos_list.length > 0) {
    pos_list.selectedIndex = pos_list.length - 1
    board.clearAnnotation()
    document.getElementById('register_moves').checked = false
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board to the current position *before* the first move
// Note that this does not only perform the move, but
// also restores any other changes on the board.
// It does NOT change the annotations.
function onBeforeFirstMove () {
  // Is the same as going the (initial) current position
  updatePositionInfoAndBoardAfterPosSelect()
  updateUI()
}

// ----------------------------------------------------------------------
// Move the board to the position *after* the previous move
// Note that this does not only perform the move, but
// also restores any other changes on the board.
// It does NOT change the annotations.
function onPreviousMove () {
  var moves_list = document.getElementById('moves_list')
  if (moves_list.selectedIndex >= 0) {
    // Got to the position before the current move, if possible
    moves_list.selectedIndex = moves_list.selectedIndex - 1
    updateBoardAfterMoveSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board to the position after the next move.
// Note that this does not only perform the move, but
// also restores any other changes on the board.
// It does NOT change the annotations.
function onNextMove () {
  var moves_list = document.getElementById('moves_list')
  if (moves_list.selectedIndex < moves_list.length - 1) {
    moves_list.selectedIndex = moves_list.selectedIndex + 1
    updateBoardAfterMoveSelect()
    updateUI()
  }
}

// ----------------------------------------------------------------------
// Move the board to the position after the last move.
// Note that this does not only perform the move, but
// also restores any other changes on the board.
// It does NOT change the annotations.
function onAfterLastMove () {
  var moves_list = document.getElementById('moves_list')
  if (moves_list.length > 0) {
    // Select lasts move
    moves_list.selectedIndex = moves_list.length - 1
    updateBoardAfterMoveSelect()
    updateUI()
  }
}

// /////////////////////////////////////////////////////////////////////////////////
// HANDLING OF MOVES

// ----------------------------------------------------------------------
// Finds out a (basic) representation of the move, based on the old position
// and the new position.
// Basic representation means: fromsquare-tosquare.
function findMoveBetweenPositions (oldPos, newPos) {
  // find out the start square of the move ...
  var start = []
  for (var square in oldPos) {
    // look for squares in the old position where a piece has been (re)moved from
    if (newPos[square] === undefined || newPos[square] === '') {
      start.push(square)
    }
  }

  // find out the end square of the move ...
  var end = []
  for (square in newPos) {
    // look for square in the new position where pieces have
    if (oldPos[square] === undefined || oldPos[square] === '' || oldPos[square] !== newPos[square]) {
      end.push(square)
    }
  }

  // Interpret the results...
  if (start.length === 0 && end.length === 1) {
    // No square became empty, but one square has a changed piece
    // (or a piece has been moved there
    // Could thus be minor promotion ...,
    // as only one square (and the same for both) has changed
    if ((oldPos[end[0]] !== undefined) &&
      oldPos[end[0]].slice(-1).match(/[QRBN]/) &&
       (end[0][1] === '1' || end[0][1] === '8')) {
      // Likely previous move was a minor promotion, check the current move list
      // for a promotion as last move
      var moves = document.getElementById('moves').value
      if (moves.length > 0 && moves.slice(-1).match(/[DTLP]/)) {
        moves = moves.slice(0, -1) + pieceBoardToUI[newPos[end[0]][1]]
        document.getElementById('moves').value = moves
      }
    }
  }
  if (start.length === 1 && end.length === 1) {
    // normal case: one start and one end square.
    return start + '-' + end
  }
  // Other cases we do not interpret. Note that start=2 and end=1
  // could be en passant, and start=2 and end=2 could be castling
  return ''
}

// ----------------------------------------------------------------------
// There are two textual representations of moves in this
// application:
//  * g1-f3 : simple (standard) representation, that is accepted
//      by a board
//  * Pf3xd5 : advanced representation (including piece, and a x
//       it is move that takes an opponents piece. This also
//       includes:
//        * o-o : castling, and similarly promotion
//        * e7-e8P : (minor) promotion)
//        * e5xd6_ep : en passent
// This function converts the advanced format to a simple move
// that can be done by the Board (which are the start and end square
// separated by a '-', and no additional info).
// Any other characters (non-alpha) that are in front of the
// move will be removed.
// And the piece annotation at the start and any info at the back
// (except for minor promotion) will be stripped.
// Castling notations will be replaced by a move by the King (e1-g1).
// Also a cross will be replaced by a dash.
function getStandardMove (move) {
  // special cases: castling
  if (move === 'o-o') return 'e8-g8'
  if (move === 'o-o-o') return 'e8-c8'
  if (move === 'O-O') return 'e1-g1'
  if (move === 'O-O-O') return 'e1-gc'

  // Split at '-' or 'x'
  var squares = move.split('-')
  if (squares.length < 2) squares = move.split('x')

  if (squares.length !== 2) return ''

  if (squares[0].length < 2 || squares[0].length > 3) return ''
  if (squares[1].length < 2 || squares[0].length > 5) return ''

  // Combine both squares (stripping prefix or postfix)
  var start = squares[0].slice(-2)
  var end = squares[1].slice(0, 2)
  if (!validSquare(start) || !validSquare(end)) return ''

  // Not a valid piece code ...
  if (squares[0].length === 3 && !validPieceUI(squares[0][0])) return ''

  // Rest could still be wrong, but is silently ignored.....
  var newMove = squares[0].slice(-2) + '-' + squares[1].slice(0, 2)

  // Assume a (minor-)promotion, keep the piece...
  if (squares[1].length === 3 && validPieceUI(squares[1][2])) {
    newMove += squares[1][2]
  }

  return newMove
}

// ----------------------------------------------------------------------
// Given a move (in any format, see above) and a board, the
// move will be done on the board, and a nice representation of the
// move (see above) will be returned in text.
//
// This function tries to be smart, i.e. King moves such as e1-e3 are
// interpreted as castling, a promotion is automatically done (to Queen),
// and in case of en-passant moves (e5-d6, with a black pawn on d5) will
// also result in removing the black pawn.
//
// This function is also used after reading a text file, and while parsing
// the text string. Therefore, it will also do some sanity-checks, and will
// skip / ignore any text that can not easily be interpreted as a move.
//
// Is only called for dummy-board.
function checkAndDoMove (board, general_move) {
  // For convenience, get the simple representation
  var move = getStandardMove(general_move)
  if (move === '') return ''

  // Get current board position
  var position = board.position()

  // Get start and end square of the move
  var squares = move.split('-')
  var start = squares[0]
  var end = squares[1]

  // Check whether the move at least contains sufficient info
  // for the squares
  if (squares[0].length < 2 || squares[1].length < 2) return ''

  var prefix = '' // Symbol for piece, none for pawn
  var move_start = squares[0].slice(-2)
  var sep = '-' // 'x' is a fiendly piece was taken
  var move_end = squares[1].slice(0, 2)
  var postfix = '' // (minor)promotion or _ep

  // If the squares are not valid squares ...
  if (!validSquare(move_start) || !validSquare(move_end)) return ''

  // At least the squares are ok, other info may/will be ignored
  // if it can not be interpreted.

  // Check for special moves.
  if (start === 'e1' && end === 'g1' && position['h1'] === 'wR') {
    // castling white short
    board.move(move)
    board.move('h1-f1')
    return 'O-O'
  } else if (start === 'e1' && end === 'c1' && position['a1'] === 'wR') {
    // castling white long
    board.move(move)
    board.move('a1-d1')
    return 'O-O-O'
  } else if (start === 'e8' && end === 'g8' && position['h8'] === 'bR') {
    // castling black short
    board.move(move)
    board.move('h8-f8')
    return 'o-o'
  } else if (start === 'e8' && end === 'c8' && position['a8'] === 'bR') {
    // castling black long
    board.move(move)
    board.move('a8-d8')
    return 'o-o-o'
  } else if (Math.abs(col2num[start[0]] - col2num[end[0]]) <= 1 &&
        start[1] === '7' && end[1] === '8' && position[start] === 'wP') {
    // white promotion
    // strip the promotion symbol, if present
    board.move(start + '-' + end.slice(0, 2))
    // Update the position, i.e. replace the pawn with the promoted piece
    position = board.position()
    if (end.length > 2 && validPieceUI(end[2])) {
      // new valid piece is given, use it
      position[end.slice(0, 2)] = 'w' + pieceUIToBoard[end[2]]
      postfix = end[2]
    } else {
      // no new (or valid) piece is given, assume Queen
      position[end.slice(0, 2)] = 'wQ'
      postfix = pieceBoardToUI['Q']
    }
    if (start[0] !== end[0]) sep = 'x' // if pawn changed column, we need a cross
    board.position(position)
  } else if (Math.abs(col2num[start[0]] - col2num[end[0]]) <= 1 &&
    start[1] === '2' && end[1] === '1' && position[start] === 'bP') {
    // black promotion
    // strip the promotion symbol, if present
    board.move(start + '-' + end.slice(0, 2))
    // Update the position, i.e. replace the pawn with the promoted piece
    position = board.position()
    if (end.length > 2 && validPieceUI(end[2])) {
      // new and valid piece is given, use it
      position[end.slice(0, 2)] = 'b' + pieceUIToBoard[end[2]]
      postfix = end[2]
    } else {
      // no new (or valid) piece is given, assume Queen
      position[end.slice(0, 2)] = 'bQ'
      postfix = pieceBoardToUI['Q']
    }
    if (start[0] !== end[0]) sep = 'x' // column change implies 'x'
    board.position(position)
  } else if (start[1] === '5' && end[1] === '6' && // from 5th to 6th row
         (Math.abs(col2num[start[0]] - col2num[end[0]]) === 1) && // moving diagonal
         position[start] === 'wP' && // it is a pawn
         position[end[0] + '5'] === 'bP' && // there is a fiendly pawn at the right loc
         (typeof position[end] === 'undefined' || position[end] === '')) { // end pos is empty
    // En passant for white
    // Move the pawn
    board.move(move)
    // Remove the other pawn ...
    position = board.position()
    delete position[end[0] + '5']
    board.position(position)
    sep = 'x'
    postfix = '_ep'
  } else if (start[1] === '4' && end[1] === '3' && // from 5th to 6th row
         (Math.abs(col2num[start[0]] - col2num[end[0]]) === 1) && // moving diagonal
         position[start] === 'bP' && // it is a pawn
         position[end[0] + '4'] === 'wP' && // there is a fiendly pawn at the right loc
         (typeof position[end] === 'undefined' || position[end] === '')) { // end pos is empty
    // En passant for black
    // Move the pawn
    board.move(move)
    // Remove the other pawn ...
    position = board.position()
    delete position[end[0] + '4']
    board.position(position)
    sep = 'x'
    postfix = '_ep'
  } else {
    // Normal case
    position = board.position()
    // Get piece that is moved
    if (position[move_start] !== undefined) {
      prefix = pieceBoardToUI[position[move_start][1]]
    }
    // Check whether we need a cross
    if (position[move_end] !== undefined && position[move_end] !== '') sep = 'x'
    // Perform move
    board.move(move)
  }
  // Return the move in nice format:
  return prefix + move_start + sep + move_end + postfix
}

// ----------------------------------------------------------------------
// Checks whether the previous move was done by white
// (if the ending square of the move is a white piece).
// Invalid moves or empty start squares also return true.
function whitePlayed (board, move) {
  var standardMove = getStandardMove(move)
  if (standardMove === '') return true
  // slice is required to remove the (minor) promotion indication
  // that may be present...
  var end = standardMove.split('-')[1].slice(0, 2)
  var position = board.position()
  // This should not happen, but could happen if
  // the moves do not match with the position ...
  // As moves can be typed, this can changed.
  if (position[end] !== undefined) return position[end][0] === 'w'

  return true
}

// ----------------------------------------------------------------------
// Determines the player who will execute
// the next move, by checking the color of
// the piece that is at the starting square
function whoMovesNext (board, move) {
  // Get a decent move representatin
  var standardMove = getStandardMove(move)
  if (standardMove === '') return ''
  // Get start square
  var start = standardMove.split('-')[0]
  var position = board.position()
  // This should not happen, but could happen if
  // the moves do not match with the position ...
  // As moves can be typed, this can changed.
  if (position[start] !== undefined) return position[start][0]

  return ''
}

// /////////////////////////////////////////////////////////////////////////////////
// UI functions ...
// /////////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------
// Resets the current stelling in the list, and updates the board,
// FEN, position_name and moves (list and field)
// It used the position_list for the information.
function updatePositionInfoAndBoardAfterPosSelect () {
  // Look for the right position ...
  var pos_list = document.getElementById('position_list')
  // If nothing selected
  if (pos_list.selectedIndex < 0) {
    if (pos_list.options.length === 0) {
      // Reset values
      document.getElementById('inputfen').value = ''
      document.getElementById('position_name').value = ''
      document.getElementById('moves').value = ''
      document.getElementById('white_pieces').value = ''
      document.getElementById('black_pieces').value = ''
      document.getElementById('moves_list').options.length = 0
      return
    }
    // Select first one..
    pos_list.selectIndex = 0
  }

  // Get all information (fen, moves) from the position_list options
  var fenandmoves = pos_list.options[pos_list.selectedIndex].value
  var fenstring = fenFromListValue(fenandmoves)
  var movesstring = movesFromListValue(fenandmoves)
  var pos_name = pos_list.options[pos_list.selectedIndex].text

  // Store the info in the UI
  document.getElementById('inputfen').value = fenstring
  document.getElementById('position_name').value = pos_name
  document.getElementById('moves').value = movesstring

  // Update the moves_list
  fillMovesList()

  // Update the board
  setPositionOnBoard(fenstring)

  // Add the pieces info (must be done *after* the board update,
  // as these use the board info)
  document.getElementById('white_pieces').value = getPiecesAsString('w')
  document.getElementById('black_pieces').value = getPiecesAsString('b')
}

// ----------------------------------------------------------------------
// Restores the right position after the currently selected move.
function updateBoardAfterMoveSelect () {
  var fenstring = ''
  var moves_list = document.getElementById('moves_list')
  if (moves_list.selectedIndex >= 0) {
    // Restore position after move
    fenstring = moves_list.options[moves_list.selectedIndex].value
  } else {
    // Restore original position
    var pos_list = document.getElementById('position_list')
    if (pos_list.selectedIndex >= 0) {
      fenstring = fenFromListValue(pos_list.options[pos_list.selectedIndex].value)
    }
  }
  setPositionOnBoard(fenstring)
}

// ----------------------------------------------------------------------
// The buttons for going through the moves and positions
// are only visible if there are positions.
// This function should be called everywhere the position
// list may be changed / updated.
function updateMoveAndPositionButtonsVisibility () {
  // Only show buttons if there are positions ...
  if (document.getElementById('position_list').options.length > 0) {
    // turn on buttons for position and moves
    document.getElementById('position_buttons').style.display = 'inline-flex'
    document.getElementById('move_buttons').style.display = 'inline-flex'
  } else {
    // turn off buttons for position and moves
    document.getElementById('position_buttons').style.display = 'none'
    document.getElementById('move_buttons').style.display = 'none'
  }
}

// ----------------------------------------------------------------------
// Updates the status line, containing position and last move
function updateStatusLine (boardfen = '') {
  // Initialize info...
  var pos_name = ''
  var zet = ''
  var move = ''

  // Look for info in the position_list..
  var pos_list = document.getElementById('position_list')
  if (pos_list.selectedIndex > -1) {
    pos_name = pos_list.options[pos_list.selectedIndex].text
  }

  // Look for info in the moves_list...
  var moves_list = document.getElementById('moves_list')
  if (moves_list.selectedIndex > -1) {
    move = moves_list.options[moves_list.selectedIndex].text
    zet = ' (na ' + stripMoveDecoration(move) + ')'
  }

  // In case no move has been played, try to find out
  // who is to move next, and display that.
  var to_move = ''
  if (moves_list.selectedIndex < 0 && moves_list.options.length > 0) {
    move = stripMoveDecoration(moves_list.options[0].text)
    to_move = whoMovesNext(board, move)
    if (to_move !== '' && validPlayer(to_move)) {
      to_move = ' [' + playerToUI[to_move] + ']'
    }
  }

  // Update the status_position
  document.getElementById('status_position').innerHTML =
    pos_name + zet + to_move

  // Update status line color
  if (boardfen === '') boardfen = board.fen()
  var pos_fen = getFENCurrentPosition(0)
  if (boardfen !== pos_fen) {
    document.getElementById('status_position').style.color = '#008CBA'
  } else {
    document.getElementById('status_position').style.color = 'black'
  }
}

// ----------------------------------------------------------------------
// Updates the recording status (color of the record button)
function updateMovesRecordingStatus () {
  var checked = document.getElementById('register_moves').checked
  if (checked) {
    document.getElementById('record').style.color = 'red'
  } else {
    document.getElementById('record').style.color = 'black'
  }
}

// ----------------------------------------------------------------------
// Updates the disabled state of some buttons ...
function updateFileButtonsStatus () {
  // Reload is only possible if a current file has been loaded
  document.getElementById('reload_file_button').disabled =
    (document.getElementById('status_filename').innerText === '[geen bestand]')
  // Get is only possible if there are positions
  document.getElementById('get_file_button').disabled =
    (document.getElementById('position_list').options.length === 0)
  // Get is only possible if there are positions
  document.getElementById('clear_file_button').disabled =
    (document.getElementById('status_filename').innerText === '[geen bestand]')
}

// ----------------------------------------------------------------------
// Updates all buttons and status info in the UI.
function updateUI () {
  updateMovesRecordingStatus()
  updateMoveAndPositionButtonsVisibility()
  updateFileButtonsStatus()
  updateStatusLine()
}

// ///////////////////////////////////////////////////////////////////////////////
// Helper functions ...
// ///////////////////////////////////////////////////////////////////////////////

// -------------------------------------------------------------------------------
// Moves List functions
//
// The moves are stored in a select item (a list of selectable items) with
// name moves_list. Below functions are for filling / manipulating this list.
// The moves_list has the moves, and puts the left (white moves) and right
// (black moves).
// The moves_list is constructed from a string with moves that is stored
// with the position.
// The moves_list consists of options, where the text is the move, and the
// value is the fenstring *after* the move.
// In that way, one can easily go to a position after a move, by simply
// updating the board to the fen.
// The fen before the first move can be obtained from the position list
// (see below).
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Adds some (none-alpha) chars to a move, to make the
// display in the dropdown list better (e.g. to make
// black moves indented).
function addMoveDecoration (move, white, seq_nr) {
  var prefix = white ? ' ' : '.......... '
  var move_nr = white ? seq_nr + '.' : ''
  return move_nr + prefix + move
}

// ---------------------------------------------------------------------------------
// Strips the (none-alpha) chars from a move.
// Moves in the moves_list have decoration, to make the
// moves of black appear to the right (see also
// addMoveDecoration()).
// This function strips that info, to get a normal
// string representation of the move.
function stripMoveDecoration (move) {
  // Remove any prefix garbage, added because of the dropdown list representation
  // All non-chars are removed.
  var start = move.search(/[a-zA-Z]/)
  if (start >= 0) move = move.slice(start, move.length)
  // Remove any remaining space...
  return move.trim()
}

// ---------------------------------------------------------------------------------
// Clean up moves in a moves_string: checks the  moves, and replaces them
// with a better representation
function cleanupMovesString (fen, moves_string) {
  // Remove punctuation
  moves_string = moves_string.replace(/\./g, ' ')
  moves_string = moves_string.replace(/,/g, ' ')
  moves_string = moves_string.replace(/;/g, ' ')

  // Split in moves
  var moves = moves_string.split(' ')

  var new_moves = ''

  // Try all moves
  dummyboard.position(fen)
  for (var i = 0; i < moves.length; ++i) {
    var move = moves[i].trim()
    var checked_move = checkAndDoMove(dummyboard, move)
    checked_move = checked_move.trim()
    if (checked_move !== '') {
      // this is a good move, maintain it.
      if (new_moves !== '') new_moves += ' '
      new_moves += checked_move
    }
  }

  return new_moves
}

// ---------------------------------------------------------------------------------
// Fills the moves list dropdown. Should be called after a change of
// position, or after a change of the moves stored with a position.
// The moves information is taken from the current position.
function fillMovesList () {
  var moves_list = document.getElementById('moves_list')
  // Clear the moves_list
  moves_list.options.length = 0

  // Look for the right position ...
  var pos_list = document.getElementById('position_list')
  if (pos_list.selectedIndex < 0) return

  // Get information (fen, moves) from the position_list options
  var fenandmoves = pos_list.options[pos_list.selectedIndex].value
  var fenstring = fenFromListValue(fenandmoves)
  var movesstring = movesFromListValue(fenandmoves)

  // Fill the moves_list
  var moves = movesstring.split(' ')
  dummyboard.position(fenstring)
  var seq_nr = 1
  for (var i = 0; i < moves.length; ++i) {
    // Perform the move
    var move = moves[i]
    move = checkAndDoMove(dummyboard, move)
    // move could be empty, if the move is not a legal move.
    if (move === '') continue
    // Position after move..
    var newfen = dummyboard.fen()
    // Determine whether white or black played
    var white = whitePlayed(dummyboard, move)
    // Add to moves_list
    moves_list.add(new Option(addMoveDecoration(move, white, seq_nr), newfen))
    if (!white) seq_nr = seq_nr + 1
  }
  moves_list.selectedIndex = -1
}

// ---------------------------------------------------------------------------------
// Adds the moves_string to the current positions moves
// Only updates the current position, still need to update the moves_list
function addMovesToCurrentPosition (append) {
  var pos_list = document.getElementById('position_list')
  var selected = pos_list.selectedIndex

  var moves_string = document.getElementById('moves').value

  // If nothing selected, we do not know where to add...
  if (selected < 0) return false

  // Get the info from the UI and the position list
  var label = pos_list.options[selected].text
  var fenstring = fenFromListValue(pos_list.options[selected].value)

  // Do we maintain the previous moves? Only if append is true
  var prev_moves = ''
  if (append) {
    prev_moves = movesFromListValue(pos_list.options[selected].value)
  }
  var new_moves =
    cleanupMovesString(fenstring, prev_moves + ' ' + moves_string)

  // Create a new option, with the current info and the new moves
  // var option = document.createElement('option')
  // option.text = label
  // option.value = fenstring + ' # ' + new_moves

  // Replace current option
  pos_list.options[selected] = new Option(label, fenstring + ' # ' + new_moves) //option
  pos_list.selectedIndex = selected

  // And select the last move
  fillMovesList()
  var moves_list = document.getElementById('moves_list')
  moves_list.selectedIndex = moves_list.length - 1

  return true
}

// ---------------------------------------------------------------------------------
// Removes the current series of moves to the current position.
// Does NOT update the label of the position, or the position itself
function removeMovesFromCurrentPosition () {
  var pos_list = document.getElementById('position_list')
  var selected = pos_list.selectedIndex
  if (selected < 0) return false

  // Reset moves square
  var label = pos_list.options[selected].text
  var fenstring = fenFromListValue(pos_list.options[selected].value)
  // var option = document.createElement('option')
  // option.text = label
  // Forget the moves ...
  // option.value = fenstring + ' # ' // + moves
  // Update option
  pos_list.options[selected] = new Option(label, fenstring + ' # ') // option
  // Keep selection
  pos_list.selectedIndex = selected

  // Clear the moves_list dropdown
  var moves_list = document.getElementById('moves_list')
  moves_list.options.length = 0

  // This will update the rest.
  // newMoveSelected()
  return true
}

// ---------------------------------------------------------------------------------
// Position List functions
//
// The list of positions are stored in a select element in html (i.e. in the 
// options of the list). The list name is 'position_list').
// For each position, the list contains as text the position name, and as
// value the fen, followed by ' # ', followed by the moves.
//
// Before the moves are added to the list, it is clean-up, any garbage
// (non-move text) is removed.
// ---------------------------------------------------------------------------------

// ---------------------------------------------------------------------------------
// Get the FEN from the stelling list value
function fenFromListValue (value) {
  if (value.split(' # ').length > 0) {
    return value.split(' # ')[0]
  } else {
    return ''
  }
}

// ---------------------------------------------------------------------------------
// Get the moves from the stelling list value
function movesFromListValue (value) {
  if (value.split(' # ').length > 1) {
    return value.split(' # ')[1]
  } else {
    return ''
  }
}

// ---------------------------------------------------------------------------------
// Searches for the first occurrence of a position with the current name...
// Returns true if found, selectedIndex will have the value of the
// found occurence.
// Returns false if not found (selectedIndex is not updated).
// Search starts at currently selected position.
function searchPositionWithName (pos_name) {
  var pos_list = document.getElementById('position_list')

  for (var i = pos_list.selectedIndex; i < pos_list.length; ++i) {
    // Skip case where there is no selection
    if (i < 0) continue
    if (pos_list.options[i].text === pos_name) {
      if (i !== pos_list.selectedIndex) {
        pos_list.selectedIndex = i
      }
      return true
    }
  }
  for (i = 0; i < pos_list.selectedIndex; ++i) {
    if (pos_list.options[i].text === pos_name) {
      pos_list.selectedIndex = i
      return true
    }
  }
  return false
}

// ---------------------------------------------------------------------------------
// Returns the FEN as it should be for the current position,
// possibly with an offset for the moves...
function getFENCurrentPosition (moveoffset) {
  // 3 Cases:
  // * no positions present or selected --> return FEN of  the board
  // * position selected, no move present or selected --> return FEN
  //   of the position
  // * position and move selected --> return FEN of the position and
  //   next move. If the current move is the last move, use the dummy
  //   board to compute the FEN ...

  var pos_list = document.getElementById('position_list')
  if (pos_list.length === 0 || pos_list.selectedIndex < 0) {
    return board.fen()
  }

  var selectedpos = pos_list.selectedIndex

  // there is at least a position selected
  var moves_list = document.getElementById('moves_list')
  if (moves_list.length === 0 || moves_list.selectedIndex + moveoffset < 0) {
    // No moves selected --> return fen of position
    return fenFromListValue(pos_list.options[selectedpos].value)
  }

  var selectedmove = moves_list.selectedIndex + moveoffset

  // there is also a move selected
  if (selectedmove < moves_list.length) {
    // Not the last move, so return the fen from the next move
    var fen = moves_list.options[selectedmove].value
    return fen.trim()
  }
  return ''
}

// ---------------------------------------------------------------------------------
// Add current position to the position list,
// based on the current position on the board.
// A name is added, and no moves are added
// If the name already exists (or no name is given),
// that position is selected, but nothing is added.
function addPositionToPosList () {
  var label = document.getElementById('position_name').value
  var pos_list = document.getElementById('position_list')
  var selected = pos_list.selectedIndex
  if (label !== '' && !searchPositionWithName(label)) {
    // If a new valid name...
    // add the position
    var fenstring = board.fen()
    // var option = document.createElement('option')
    // option.text = label
    // option.value = fenstring
    // Add after current
    // pos_list.add(option, pos_list.selectedIndex + 1)
    pos_list.add(new Option(label, fenstring), pos_list.selectedIndex + 1)
    pos_list.selectedIndex = pos_list.selectedIndex + 1
    // updateMoveAndPositionButtonsVisibility()
    // newPositionSelected()
    return true
  } else {
    // No valid name, thus reset the selectedIndex, which may have been 
    // changed by the searchPositionWithName
    pos_list.selectedIndex = selected
  }
  return false
}

// ---------------------------------------------------------------------------------
// Removes the position with name (with moves) from the list
function removePositionFromPosList () {
  var label = document.getElementById('position_name').value

  if (searchPositionWithName(label)) {
    // And the item is being removed.
    var pos_list = document.getElementById('position_list')
    var index = pos_list.selectedIndex
    pos_list.remove(pos_list.selectedIndex)
    // Update selected index...
    if (index < pos_list.options.length) {
      pos_list.selectedIndex = index
    } else {
      pos_list.selectedIndex = index - 1
    }
    // updateMoveAndPositionButtonsVisibility()
    // Update the board to the new position.
    // newPositionSelected()
    return true
  }
  return false
}

// ---------------------------------------------------------------------------------
// Update (change) current position to the position list,
// based on the current position on the board, and the
// name specified, and the current moves.
// (i.e. moves are not touched!)
function updatePositionInPosList () {
  var pos_list = document.getElementById('position_list')

  // No options, or none selected...
  if (pos_list.options.length === 0 ||
    pos_list.selectedIndex < 0) {
    return false
  }

  // Updates can also change the name, thus we prevent duplicate names
  var newname = document.getElementById('position_name').value
  // Do not process empty names...
  if (newname === '') return false

  // If the name did not change, or the new name does not exist..
  if (pos_list.options[pos_list.selectedIndex].text === newname ||
    !searchPositionWithName(newname)) {
    // var fenstring = document.getElementById('inputfen').value
    var fenstring = board.fen()
    // Do not update the moves, as they have their own way
    // to be updated. This only changes the (initial) position
    // var moves = document.getElementById('moves').value
    var selected = pos_list.selectedIndex
    var prev_moves = movesFromListValue(pos_list.options[selected].value)
    // var option = document.createElement('option')
    // option.text = newname
    // option.value = fenstring  + ' # ' + prev_moves
    pos_list.options[selected] = new Option(newname, fenstring + ' # ' + prev_moves)
    pos_list.selectedIndex = selected
    // newPositionSelected()
    return true
  }
  return false
}

// /////////////////////////////////////////////////////////////////////////////////
// File loading and saving
// /////////////////////////////////////////////////////////////////////////////////

// Event function for opening a file.
function onOpenFileEvent (event) {
  // Get the files
  var files = event.target.files
  // In case of cancel, files.length === 0
  if (files.length > 0) {
    // Open them (only 1)
    openFile(files)

    // Update status etc.
    document.getElementById('register_moves').checked = false
    board.clearAnnotation()
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
}

// ---------------------------------------------------------------------------------
// Event function for reloading the current file.
function onReloadFile () {
  // Reloads the current file
  var loadfile = document.getElementById('loadfile')
  openFile(loadfile.files)

  // Update status etc.
  document.getElementById('register_moves').checked = false
  board.clearAnnotation()
  updatePositionInfoAndBoardAfterPosSelect()
  updateUI()
}

// ---------------------------------------------------------------------------------
// Clears the position list, filename etc.
function onClearFile () {
  // Reset the current loaded file name
  var loadfile = document.getElementById('loadfile')
  loadfile.value = ''

  // Open those, which will reset position list, move list and status.
  openFile(loadfile.files)

  // Update status etc.
  document.getElementById('register_moves').checked = false
  document.getElementById('status_filename').innerText = '[geen bestand]'
  board.clearAnnotation()
  updatePositionInfoAndBoardAfterPosSelect()
  updateUI()
}

// ---------------------------------------------------------------------------------
// Does the parsing of the file contents.
function readV10File (lines) {
  var pos_list = document.getElementById('position_list')
  for (var i = 0; i < lines.length; ++i) {
    var line = lines[i].trim()
    // Only process non-empty lines that are not commented
    if ((line !== '') && (line[0] !== '#')) {
      var keyvalue = line.split(' | ')
      // Add the FEN and moves in one string. These will be interpreted by newPositionSelected()!
      if (keyvalue.length === 2) {
        pos_list.add(new Option(keyvalue[0].trim(), keyvalue[1].trim()))
        // addOptionToList(pos_list, keyvalue[0].trim(), keyvalue[1].trim())
      }
    }
  }
}

// ---------------------------------------------------------------------------------
// Reads positions and moves from a file.
// Current format:
//   <stelling-name> | <FEN> # <moves>
// on one line.
// Lines starting with a # are assumed to be comments,
// the # and moves do not need to be present (moves are optional).
// Input files are text files, and can thus be edited manually.
// There is not much checking of the file format, and contents, so
// in case of errors in the file, you may have a problem.
function openFile (input) {
  // var input = event.target

  var pos_list = document.getElementById('position_list')
  pos_list.options.length = 0
  var moves_list = document.getElementById('moves_list')
  moves_list.options.length = 0

  // No file case ...
  if (input.length <= 0 || input[0] === '') {
    document.getElementById('filename').innerText = ''
    document.getElementById('status_filename').innerText = '[geen bestand]'
    return
  }

  var reader = new FileReader()
  reader.onload = function () {
    var text = reader.result
    var lines = text.split('\n')
    // Test the first line for a version number
    if (lines[0].match('# ddm-1.0')) {
      // First version with a header
      lines.shift()
      readV10File(lines)
    } else {
      // Zero version: no header
      readV10File(lines)
    }
    // if positions read ...
    if (document.getElementById('position_list').options.length > 0) {
      // Start with first stelling after reading a file.
      document.getElementById('position_list').selectedIndex = 0
    }
    // moves and position buttons are only visible if there are positions
    updatePositionInfoAndBoardAfterPosSelect()
    updateUI()
  }
  // Things that are put here may be executed *before* the reader has finished (?!)
  reader.readAsText(input[0])
  document.getElementById('filename').innerText = input[0].name
  document.getElementById('status_filename').innerText = input[0].name + ':'
}

// For downloading a file...
// From the internet ...
function downloadFile (filename, text) {
  var element = document.createElement('a')
  element.setAttribute('href', 'data:text/plain;charset=utf-8,' + encodeURIComponent(text))
  element.setAttribute('download', filename)

  element.style.display = 'none'
  document.body.appendChild(element)

  element.click()

  document.body.removeChild(element)
}

// Produces the text that represents the current list of stellingen
// and allows the user to save this text to a file via a download
// (it is not possible to directly save to a local drive via JavaScript).
function onGetFENFileText () {
  var pos_list = document.getElementById('position_list')
  var text = '# ddb-1.0: version 1.0 of digidemobord files\n'
  for (var i = 0; i < pos_list.options.length; ++i) {
    var option = pos_list.options[i]
    text += option.text + ' | ' + option.value + '\n'
  }
  document.getElementById('fenfile').value = text
  document.getElementById('fenfile').select()

  /* Copy the text inside the text field */
  document.execCommand('copy')

  // Allow user to download the file.
  downloadFile('stellingen.txt', text)
}

// /////////////////////////////////////////////////////////////////////////////////
// Simple Helper Functions for Pieces and Squares
// These functions also are relevant for internationalization
// /////////////////////////////////////////////////////////////////////////////////

// Copied from chessboard
function isString (s) {
  return typeof s === 'string'
}

// Copied from chessboard
function validSquare (square) {
  return isString(square) && square.search(/^[a-h][1-8]$/) !== -1
}

// Copied from chessboard
function validPieceCode (code) {
  return isString(code) && code.search(/^[bw][KQRNBP]$/) !== -1
}

// Conversion from column name to an index (number)
var col2num = {
  'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6, 'g': 7, 'h': 8
}

// Converts the character for a piece on the board (English) to a
// character for the UI (which is Dutch)
var pieceBoardToUI = {
  'K': 'K', 'Q': 'D', 'B': 'L', 'N': 'P', 'R': 'T', 'P': ''
}

// Converts the character for a player on the board (English) to a
// character for the UI (which is Dutch)
var playerToUI = {
  'w': 'W', 'b': 'Z'
}

// Converts the character for a piece in the UI (Dutch) to a
// character for the board (which is English)
var pieceUIToBoard = {
  'K': 'K', 'D': 'Q', 'L': 'B', 'P': 'N', 'T': 'R', '': 'P'
}

// Checks whether the piece is a valid piece in the UI
// (by checking the keys of pieceUIToBoard)
function validPieceUI (piece) {
  for (var p in pieceUIToBoard) {
    if (p === piece) return true
  }
  return false
}

// Returns  whether a player is valid (in Board terminology)
// (by checking the keys of playerToUI)
function validPlayer (player) {
  for (var p in playerToUI) {
    if (p === player) return true
  }
  return false
}

// ----------------------------------------------------------------------
// Returns all pieces of a certain color (either 'w' or 'b')
// that are currently on the board. These are returned in a text string,
// to be added to the user interface.
//
// E.g. "a2 b2 c2 Dd1 Ke2 Th1"
function getPiecesAsString (color) {
  var position = board.position()
  var pieces_string = ''
  // For all squares...
  for (var square in position) {
    var piece = position[square]
    // If right color
    if (piece[0] === color) {
      // Add the piece
      pieces_string += ' ' + pieceBoardToUI[piece[1]] + square
    }
  }
  return pieces_string.trim()
}

// ----------------------------------------------------------------------
// Global key events
const body_node = document.getElementById('body')
body_node.addEventListener('keypress', onKeyPress)

// Events listener (to ENter) for inputfen --> set board to fen
document.getElementById('inputfen').addEventListener('keypress', onKeyPressInputFEN)

// Events listener (to Enter) for stellingname --> adds stelling to list
document.getElementById('position_name').addEventListener('keypress', onKeyPressPosition)

// Events listener for moves --> prevents that any typing in this square leads
// to global event interpretation
document.getElementById('moves').addEventListener('keypress', onKeyPressMoves)

// Events listener for white and black pieces --> prevents that any typing in this
// leads to global event interpretation
document.getElementById('white_pieces').addEventListener('keypress', onKeyPressWhitePieces)
document.getElementById('black_pieces').addEventListener('keypress', onKeyPressBlackPieces)

// keypress does not seem to work for Escape?!
document.getElementById('search_status_position').addEventListener('keyup', onKeyUpSearchPosition)
document.getElementById('match_list').addEventListener('keyup', onKeyUpSearchPosition)
// And keypress required to prevent passing the keypress to other elements..
document.getElementById('search_status_position').addEventListener('keypress', onKeyPressSearchPosition)

// Set the state and visibility of relevant UI elements.
updateUI()
